@@ -29,6 +29,7 @@
 #include <drm_fourcc.h>
 #include <wlr/interfaces/wlr_buffer.h>
 #include "buffer.h"
+#include "common/box.h"
 #include "common/mem.h"
 
 static const struct wlr_buffer_impl data_buffer_impl;
@@ -44,14 +45,12 @@ static void
 data_buffer_destroy(struct wlr_buffer *wlr_buffer)
 {
 	struct lab_data_buffer *buffer = data_buffer_from_buffer(wlr_buffer);
-	if (!buffer->free_on_destroy) {
-		free(buffer);
-		return;
-	}
 	if (buffer->cairo) {
-		cairo_surface_t *surf = cairo_get_target(buffer->cairo);
 		cairo_destroy(buffer->cairo);
-		cairo_surface_destroy(surf);
+	}
+	if (buffer->surface) {
+		/* this also frees buffer->data */
+		cairo_surface_destroy(buffer->surface);
 	} else if (buffer->data) {
 		free(buffer->data);
 		buffer->data = NULL;
@@ -85,19 +84,35 @@ static const struct wlr_buffer_impl data_buffer_impl = {
 };
 
 struct lab_data_buffer *
-buffer_create_cairo(uint32_t width, uint32_t height, float scale,
-	bool free_on_destroy)
+buffer_adopt_cairo_surface(cairo_surface_t *surface)
 {
-	struct lab_data_buffer *buffer = znew(*buffer);
-	buffer->unscaled_width = width;
-	buffer->unscaled_height = height;
-	width *= scale;
-	height *= scale;
+	assert(cairo_surface_get_type(surface) == CAIRO_SURFACE_TYPE_IMAGE);
+	assert(cairo_image_surface_get_format(surface) == CAIRO_FORMAT_ARGB32);
 
-	/* Allocate the buffer with the scaled size */
+	int width = cairo_image_surface_get_width(surface);
+	int height = cairo_image_surface_get_height(surface);
+
+	struct lab_data_buffer *buffer = znew(*buffer);
 	wlr_buffer_init(&buffer->base, &data_buffer_impl, width, height);
-	cairo_surface_t *surf =
-		cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
+
+	buffer->surface = surface;
+	buffer->data = cairo_image_surface_get_data(buffer->surface);
+	buffer->format = DRM_FORMAT_ARGB8888;
+	buffer->stride = cairo_image_surface_get_stride(buffer->surface);
+	buffer->logical_width = width;
+	buffer->logical_height = height;
+
+	return buffer;
+}
+
+struct lab_data_buffer *
+buffer_create_cairo(uint32_t logical_width, uint32_t logical_height, float scale)
+{
+	/* Create an image surface with the scaled size */
+	cairo_surface_t *surface =
+		cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
+			lroundf(logical_width * scale),
+			lroundf(logical_height * scale));
 
 	/**
 	 * Tell cairo about the device scale so we can keep drawing in unscaled
@@ -106,34 +121,82 @@ buffer_create_cairo(uint32_t width, uint32_t height, float scale,
 	 *
 	 * For a more complete explanation see PR #389
 	 */
-	cairo_surface_set_device_scale(surf, scale, scale);
+	cairo_surface_set_device_scale(surface, scale, scale);
 
-	buffer->cairo = cairo_create(surf);
-	buffer->data = cairo_image_surface_get_data(surf);
-	buffer->format = DRM_FORMAT_ARGB8888;
-	buffer->stride = cairo_image_surface_get_stride(surf);
-	buffer->free_on_destroy = free_on_destroy;
+	/*
+	 * Adopt the image surface into a buffer, set the correct
+	 * logical size, and create a cairo context for drawing
+	 */
+	struct lab_data_buffer *buffer = buffer_adopt_cairo_surface(surface);
+	buffer->logical_width = logical_width;
+	buffer->logical_height = logical_height;
+	buffer->cairo = cairo_create(surface);
 
-	if (!buffer->data) {
-		cairo_destroy(buffer->cairo);
-		cairo_surface_destroy(surf);
-		free(buffer);
-		buffer = NULL;
-	}
 	return buffer;
 }
 
 struct lab_data_buffer *
-buffer_create_wrap(void *pixel_data, uint32_t width, uint32_t height,
-	uint32_t stride, bool free_on_destroy)
+buffer_convert_cairo_surface_for_icon(cairo_surface_t *surface,
+		uint32_t icon_size, float scale)
+{
+	assert(cairo_surface_get_type(surface) == CAIRO_SURFACE_TYPE_IMAGE);
+
+	/*
+	 * Compute logical size for display and decide whether we can
+	 * use the image data directly (fast path). Requirements are:
+	 *
+	 *  - The pixel format must be ARGB32.
+	 *  - The image must not be so large as to need downsampling by
+	 *    more than 2x when displayed at the target scale. wlr_scene
+	 *    uses linear interpolation without pixel averaging, which
+	 *    starts to skip samples if downsampling more than 2x,
+	 *    resulting in a grainy look.
+	 */
+	int width = cairo_image_surface_get_width(surface);
+	int height = cairo_image_surface_get_height(surface);
+	struct wlr_box logical =
+		box_fit_within(width, height, icon_size, icon_size);
+	struct lab_data_buffer *buffer;
+
+	if (cairo_image_surface_get_format(surface) == CAIRO_FORMAT_ARGB32
+			&& width <= 2 * logical.width * scale
+			&& height <= 2 * logical.height * scale) {
+		buffer = buffer_adopt_cairo_surface(surface);
+		/* set logical size for display */
+		buffer->logical_width = logical.width;
+		buffer->logical_height = logical.height;
+	} else {
+		/* convert to ARGB32 and scale for display (slow path) */
+		buffer = buffer_create_cairo(logical.width,
+			logical.height, scale);
+
+		cairo_t *cairo = buffer->cairo;
+		cairo_scale(cairo, (double)logical.width / width,
+			(double)logical.height / height);
+		cairo_set_source_surface(cairo, surface, 0, 0);
+		cairo_pattern_set_filter(cairo_get_source(cairo),
+			CAIRO_FILTER_GOOD);
+		cairo_paint(cairo);
+
+		/* ensure pixel data is updated */
+		cairo_surface_flush(buffer->surface);
+		/* destroy original surface */
+		cairo_surface_destroy(surface);
+	}
+
+	return buffer;
+}
+
+struct lab_data_buffer *
+buffer_create_from_data(void *pixel_data, uint32_t width, uint32_t height,
+		uint32_t stride)
 {
 	struct lab_data_buffer *buffer = znew(*buffer);
 	wlr_buffer_init(&buffer->base, &data_buffer_impl, width, height);
-	buffer->unscaled_width = width;
-	buffer->unscaled_height = height;
+	buffer->logical_width = width;
+	buffer->logical_height = height;
 	buffer->data = pixel_data;
 	buffer->format = DRM_FORMAT_ARGB8888;
 	buffer->stride = stride;
-	buffer->free_on_destroy = free_on_destroy;
 	return buffer;
 }
