@@ -19,15 +19,19 @@
 #include "idle.h"
 #include "action.h"
 
-static bool
-tool_supports_absolute_motion(struct wlr_tablet_tool *tool)
+static enum motion
+tool_motion_mode(enum motion motion, struct wlr_tablet_tool *tool)
 {
+	/*
+	 * Absolute positioning doesn't make sense
+	 * for tablet mouses and lenses.
+	 */
 	switch (tool->type) {
 	case WLR_TABLET_TOOL_TYPE_MOUSE:
 	case WLR_TABLET_TOOL_TYPE_LENS:
-		return false;
+		return LAB_TABLET_MOTION_RELATIVE;
 	default:
-		return true;
+		return motion;
 	}
 }
 
@@ -83,30 +87,86 @@ adjust_for_rotation(enum rotation rotation, double *x, double *y)
 	}
 }
 
+static void
+adjust_for_rotation_relative(enum rotation rotation, double *dx, double *dy)
+{
+	double tmp;
+	switch (rotation) {
+	case LAB_ROTATE_NONE:
+		break;
+	case LAB_ROTATE_90:
+		tmp = *dx;
+		*dx = -*dy;
+		*dy = tmp;
+		break;
+	case LAB_ROTATE_180:
+		*dx = -*dx;
+		*dy = -*dy;
+		break;
+	case LAB_ROTATE_270:
+		tmp = *dx;
+		*dx = *dy;
+		*dy = -tmp;
+		break;
+	}
+}
+
+static void
+adjust_for_motion_sensitivity(double motion_sensitivity, double *dx, double *dy)
+{
+	*dx = *dx * motion_sensitivity;
+	*dy = *dy * motion_sensitivity;
+}
+
 static struct wlr_surface*
-tablet_get_coords(struct drawing_tablet *tablet, double *x, double *y)
+tablet_get_coords(struct drawing_tablet *tablet, double *x, double *y, double *dx, double *dy)
 {
 	*x = tablet->x;
 	*y = tablet->y;
+	*dx = tablet->dx;
+	*dy = tablet->dy;
+	adjust_for_rotation(rc.tablet.rotation, x, y);
 	adjust_for_tablet_area(tablet->tablet->width_mm, tablet->tablet->height_mm,
 		rc.tablet.box, x, y);
-	adjust_for_rotation(rc.tablet.rotation, x, y);
+	adjust_for_rotation_relative(rc.tablet.rotation, dx, dy);
+	adjust_for_motion_sensitivity(rc.tablet_tool.relative_motion_sensitivity, dx, dy);
 
+	/*
+	 * Do not return a surface when mouse emulation is enforced. Not
+	 * having a surface or tablet tool (see handle_tablet_tool_proximity())
+	 * will trigger the fallback to cursor move/button emulation in the
+	 * tablet signal handlers.
+	 */
 	if (rc.tablet.force_mouse_emulation
 			|| !tablet->tablet_v2) {
 		return NULL;
 	}
 
-	/* Convert coordinates: first [0, 1] => layout, then layout => surface */
-	double lx, ly;
-	wlr_cursor_absolute_to_layout_coords(tablet->seat->cursor,
-		tablet->wlr_input_device, *x, *y, &lx, &ly);
+	/* convert coordinates: first [0, 1] => layout, then layout => surface */
+
+	/* initialize here to avoid a maybe-uninitialized compiler warning */
+	double lx = -1, ly = -1;
+	switch (tablet->motion_mode) {
+	case LAB_TABLET_MOTION_ABSOLUTE:
+		wlr_cursor_absolute_to_layout_coords(tablet->seat->cursor,
+			tablet->wlr_input_device, *x, *y, &lx, &ly);
+		break;
+	case LAB_TABLET_MOTION_RELATIVE:
+		/*
+		 * Deltas dx,dy will be directly passed into wlr_cursor_move,
+		 * so we can add those directly here to determine our future
+		 * position.
+		 */
+		lx = tablet->seat->cursor->x + *dx;
+		ly = tablet->seat->cursor->y + *dy;
+		break;
+	}
 
 	double sx, sy;
 	struct wlr_scene_node *node =
 		wlr_scene_node_at(&tablet->seat->server->scene->tree.node, lx, ly, &sx, &sy);
 
-	/* Find the surface and return it if it accepts tablet events */
+	/* find the surface and return it if it accepts tablet events */
 	struct wlr_surface *surface = lab_wlr_surface_from_node(node);
 
 	if (surface && !wlr_surface_accepts_tablet_v2(tablet->tablet_v2, surface)) {
@@ -117,7 +177,8 @@ tablet_get_coords(struct drawing_tablet *tablet, double *x, double *y)
 
 static void
 notify_motion(struct drawing_tablet *tablet, struct drawing_tablet_tool *tool,
-		struct wlr_surface *surface, double x, double y, uint32_t time)
+		struct wlr_surface *surface, double x, double y, double dx, double dy,
+		uint32_t time)
 {
 	idle_manager_notify_activity(tool->seat->seat);
 
@@ -129,8 +190,16 @@ notify_motion(struct drawing_tablet *tablet, struct drawing_tablet_tool *tool,
 			tablet->tablet_v2, surface);
 	}
 
-	wlr_cursor_warp_absolute(tablet->seat->cursor,
-		tablet->wlr_input_device, x, y);
+	switch (tablet->motion_mode) {
+	case LAB_TABLET_MOTION_ABSOLUTE:
+		wlr_cursor_warp_absolute(tablet->seat->cursor,
+			tablet->wlr_input_device, x, y);
+		break;
+	case LAB_TABLET_MOTION_RELATIVE:
+		wlr_cursor_move(tablet->seat->cursor,
+			tablet->wlr_input_device, dx, dy);
+		break;
+	}
 
 	double sx, sy;
 	bool notify = cursor_process_motion(tablet->seat->server, time, &sx, &sy);
@@ -173,32 +242,47 @@ notify_motion(struct drawing_tablet *tablet, struct drawing_tablet_tool *tool,
 }
 
 static void
-handle_proximity(struct wl_listener *listener, void *data)
+handle_tablet_tool_proximity(struct wl_listener *listener, void *data)
 {
 	struct wlr_tablet_tool_proximity_event *ev = data;
 	struct drawing_tablet *tablet = ev->tablet->data;
 	struct drawing_tablet_tool *tool = ev->tool->data;
-
-	if (!tool_supports_absolute_motion(ev->tool)) {
-		if (ev->state == WLR_TABLET_TOOL_PROXIMITY_IN) {
-			wlr_log(WLR_INFO, "ignoring not supporting tablet tool");
-		}
+	if (!tablet) {
+		wlr_log(WLR_DEBUG, "tool proximity event before tablet create");
 		return;
 	}
 
+	if (ev->state == WLR_TABLET_TOOL_PROXIMITY_IN) {
+		tablet->motion_mode =
+			tool_motion_mode(rc.tablet_tool.motion, ev->tool);
+	}
+
+	/*
+	 * Reset relative coordinates, we don't want to move the
+	 * cursor on proximity-in for relative positioning.
+	 */
+	tablet->dx = 0;
+	tablet->dy = 0;
+
 	tablet->x = ev->x;
 	tablet->y = ev->y;
 
-	double x, y;
-	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y);
+	double x, y, dx, dy;
+	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y, &dx, &dy);
 
+	/*
+	 * Do not attempt to create a tablet tool when mouse emulation is
+	 * enforced. Not having a tool or tablet capable surface will trigger
+	 * the fallback to cursor move/button emulation in the tablet signal
+	 * handlers.
+	 */
 	if (!rc.tablet.force_mouse_emulation
 			&& tablet->seat->server->tablet_manager && !tool) {
 		/*
 		 * Unfortunately `wlr_tool` is only present in the events, so
 		 * use proximity for creating a `wlr_tablet_v2_tablet_tool`.
 		 */
-		tablet_tool_init(tablet->seat, ev->tool);
+		tablet_tool_create(tablet->seat, ev->tool);
 	}
 
 	/*
@@ -207,7 +291,7 @@ handle_proximity(struct wl_listener *listener, void *data)
 	 */
 	if (tool && surface) {
 		if (tool->tool_v2 && ev->state == WLR_TABLET_TOOL_PROXIMITY_IN) {
-			notify_motion(tablet, tool, surface, x, y, ev->time_msec);
+			notify_motion(tablet, tool, surface, x, y, dx, dy, ev->time_msec);
 		}
 		if (tool->tool_v2 && ev->state == WLR_TABLET_TOOL_PROXIMITY_OUT) {
 			wlr_tablet_v2_tablet_tool_notify_proximity_out(tool->tool_v2);
@@ -218,21 +302,32 @@ handle_proximity(struct wl_listener *listener, void *data)
 static bool is_down_mouse_emulation = false;
 
 static void
-handle_axis(struct wl_listener *listener, void *data)
+handle_tablet_tool_axis(struct wl_listener *listener, void *data)
 {
 	struct wlr_tablet_tool_axis_event *ev = data;
 	struct drawing_tablet *tablet = ev->tablet->data;
 	struct drawing_tablet_tool *tool = ev->tool->data;
-
-	if (!tool_supports_absolute_motion(ev->tool)) {
+	if (!tablet) {
+		wlr_log(WLR_DEBUG, "tool axis event before tablet create");
 		return;
 	}
 
+	/*
+	 * Reset relative coordinates. If those axes aren't updated,
+	 * the delta is zero.
+	 */
+	tablet->dx = 0;
+	tablet->dy = 0;
+	tablet->tilt_x = 0;
+	tablet->tilt_y = 0;
+
 	if (ev->updated_axes & WLR_TABLET_TOOL_AXIS_X) {
 		tablet->x = ev->x;
+		tablet->dx = ev->dx;
 	}
 	if (ev->updated_axes & WLR_TABLET_TOOL_AXIS_Y) {
 		tablet->y = ev->y;
+		tablet->dy = ev->dy;
 	}
 	if (ev->updated_axes & WLR_TABLET_TOOL_AXIS_DISTANCE) {
 		tablet->distance = ev->distance;
@@ -256,8 +351,8 @@ handle_axis(struct wl_listener *listener, void *data)
 		tablet->wheel_delta = ev->wheel_delta;
 	}
 
-	double x, y;
-	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y);
+	double x, y, dx, dy;
+	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y, &dx, &dy);
 
 	/*
 	 * We are sending tablet notifications on the following conditions:
@@ -274,7 +369,7 @@ handle_axis(struct wl_listener *listener, void *data)
 			&& tablet->seat->server->input_mode == LAB_INPUT_STATE_PASSTHROUGH)
 			|| wlr_tablet_tool_v2_has_implicit_grab(tool->tool_v2))) {
 		/* motion seems to be supported by all tools */
-		notify_motion(tablet, tool, surface, x, y, ev->time_msec);
+		notify_motion(tablet, tool, surface, x, y, dx, dy, ev->time_msec);
 
 		/* notify about other axis based on tool capabilities */
 		if (ev->tool->distance) {
@@ -299,7 +394,7 @@ handle_axis(struct wl_listener *listener, void *data)
 			 */
 			double tilt_x = tablet->tilt_x;
 			double tilt_y = tablet->tilt_y;
-			adjust_for_rotation(rc.tablet.rotation, &tilt_x, &tilt_y);
+			adjust_for_rotation_relative(rc.tablet.rotation, &tilt_x, &tilt_y);
 
 			wlr_tablet_v2_tablet_tool_notify_tilt(tool->tool_v2,
 				tilt_x, tilt_y);
@@ -322,8 +417,19 @@ handle_axis(struct wl_listener *listener, void *data)
 				wlr_tablet_v2_tablet_tool_notify_proximity_out(
 					tool->tool_v2);
 			}
-			cursor_emulate_move_absolute(tablet->seat, &ev->tablet->base,
-				x, y, ev->time_msec);
+
+			switch (tablet->motion_mode) {
+			case LAB_TABLET_MOTION_ABSOLUTE:
+				cursor_emulate_move_absolute(tablet->seat,
+					&ev->tablet->base,
+					x, y, ev->time_msec);
+				break;
+			case LAB_TABLET_MOTION_RELATIVE:
+				cursor_emulate_move(tablet->seat,
+					&ev->tablet->base,
+					dx, dy, ev->time_msec);
+				break;
+			}
 		}
 	}
 }
@@ -374,14 +480,18 @@ seat_pointer_end_grab(struct drawing_tablet_tool *tool,
 }
 
 static void
-handle_tip(struct wl_listener *listener, void *data)
+handle_tablet_tool_tip(struct wl_listener *listener, void *data)
 {
 	struct wlr_tablet_tool_tip_event *ev = data;
 	struct drawing_tablet *tablet = ev->tablet->data;
 	struct drawing_tablet_tool *tool = ev->tool->data;
+	if (!tablet) {
+		wlr_log(WLR_DEBUG, "tool tip event before tablet create");
+		return;
+	}
 
-	double x, y;
-	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y);
+	double x, y, dx, dy;
+	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y, &dx, &dy);
 
 	uint32_t button = tablet_get_mapped_button(BTN_TOOL_PEN);
 
@@ -418,7 +528,7 @@ handle_tip(struct wl_listener *listener, void *data)
 				wlr_tablet_v2_tablet_tool_notify_up(tool->tool_v2);
 			}
 
-			bool exit_interactive = cursor_finish_button_release(tool->seat);
+			bool exit_interactive = cursor_finish_button_release(tool->seat, BTN_LEFT);
 			if (exit_interactive && surface && tool->tool_v2->focused_surface) {
 				/*
 				 * Re-enter the surface after a resize/move to ensure
@@ -441,22 +551,26 @@ handle_tip(struct wl_listener *listener, void *data)
 			cursor_emulate_button(tablet->seat,
 				button,
 				ev->state == WLR_TABLET_TOOL_TIP_DOWN
-					? WLR_BUTTON_PRESSED
-					: WLR_BUTTON_RELEASED,
+					? WL_POINTER_BUTTON_STATE_PRESSED
+					: WL_POINTER_BUTTON_STATE_RELEASED,
 				ev->time_msec);
 		}
 	}
 }
 
 static void
-handle_button(struct wl_listener *listener, void *data)
+handle_tablet_tool_button(struct wl_listener *listener, void *data)
 {
 	struct wlr_tablet_tool_button_event *ev = data;
 	struct drawing_tablet *tablet = ev->tablet->data;
 	struct drawing_tablet_tool *tool = ev->tool->data;
+	if (!tablet) {
+		wlr_log(WLR_DEBUG, "tool button event before tablet create");
+		return;
+	}
 
-	double x, y;
-	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y);
+	double x, y, dx, dy;
+	struct wlr_surface *surface = tablet_get_coords(tablet, &x, &y, &dx, &dy);
 
 	uint32_t button = tablet_get_mapped_button(ev->button);
 
@@ -479,7 +593,7 @@ handle_button(struct wl_listener *listener, void *data)
 						&& mousebind->button == button
 						&& mousebind->context == LAB_SSD_CLIENT) {
 					actions_run(view, tool->seat->server,
-						&mousebind->actions, 0);
+						&mousebind->actions, NULL);
 				}
 			}
 		}
@@ -500,7 +614,11 @@ handle_button(struct wl_listener *listener, void *data)
 	} else {
 		if (button) {
 			is_down_mouse_emulation = ev->state == WLR_BUTTON_PRESSED;
-			cursor_emulate_button(tablet->seat, button, ev->state, ev->time_msec);
+			cursor_emulate_button(tablet->seat, button,
+				ev->state == WLR_BUTTON_PRESSED
+					? WL_POINTER_BUTTON_STATE_PRESSED
+					: WL_POINTER_BUTTON_STATE_RELEASED,
+				ev->time_msec);
 		}
 	}
 }
@@ -514,16 +632,12 @@ handle_destroy(struct wl_listener *listener, void *data)
 	wl_list_remove(&tablet->link);
 	tablet_pad_attach_tablet(tablet->seat);
 
-	wl_list_remove(&tablet->handlers.tip.link);
-	wl_list_remove(&tablet->handlers.button.link);
-	wl_list_remove(&tablet->handlers.proximity.link);
-	wl_list_remove(&tablet->handlers.axis.link);
 	wl_list_remove(&tablet->handlers.destroy.link);
 	free(tablet);
 }
 
 void
-tablet_init(struct seat *seat, struct wlr_input_device *wlr_device)
+tablet_create(struct seat *seat, struct wlr_input_device *wlr_device)
 {
 	wlr_log(WLR_DEBUG, "setting up tablet");
 	struct drawing_tablet *tablet = znew(*tablet);
@@ -546,12 +660,26 @@ tablet_init(struct seat *seat, struct wlr_input_device *wlr_device)
 	tablet->wheel_delta = 0.0;
 	wlr_log(WLR_INFO, "tablet dimensions: %.2fmm x %.2fmm",
 		tablet->tablet->width_mm, tablet->tablet->height_mm);
-	CONNECT_SIGNAL(tablet->tablet, &tablet->handlers, axis);
-	CONNECT_SIGNAL(tablet->tablet, &tablet->handlers, proximity);
-	CONNECT_SIGNAL(tablet->tablet, &tablet->handlers, tip);
-	CONNECT_SIGNAL(tablet->tablet, &tablet->handlers, button);
 	CONNECT_SIGNAL(wlr_device, &tablet->handlers, destroy);
 
 	wl_list_insert(&seat->tablets, &tablet->link);
 	tablet_pad_attach_tablet(tablet->seat);
 }
+
+void
+tablet_init(struct seat *seat)
+{
+	CONNECT_SIGNAL(seat->cursor, seat, tablet_tool_axis);
+	CONNECT_SIGNAL(seat->cursor, seat, tablet_tool_proximity);
+	CONNECT_SIGNAL(seat->cursor, seat, tablet_tool_tip);
+	CONNECT_SIGNAL(seat->cursor, seat, tablet_tool_button);
+}
+
+void
+tablet_finish(struct seat *seat)
+{
+	wl_list_remove(&seat->tablet_tool_axis.link);
+	wl_list_remove(&seat->tablet_tool_proximity.link);
+	wl_list_remove(&seat->tablet_tool_tip.link);
+	wl_list_remove(&seat->tablet_tool_button.link);
+}
