@@ -225,6 +225,7 @@ render_osd(struct server *server, cairo_t *cairo, int w, int h,
 	/* Set up text rendering */
 	set_cairo_color(cairo, theme->osd_label_text_color);
 	PangoLayout *layout = pango_cairo_create_layout(cairo);
+	pango_context_set_round_glyph_positions(pango_layout_get_context(layout), false);
 	pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_END);
 
 	PangoFontDescription *desc = font_to_pango_desc(&rc.font_osd);
@@ -330,7 +331,7 @@ display_osd(struct output *output, struct wl_array *views)
 	struct server *server = output->server;
 	struct theme *theme = server->theme;
 	bool show_workspace = wl_list_length(&rc.workspace_config.workspaces) > 1;
-	const char *workspace_name = server->workspace_current->name;
+	const char *workspace_name = server->workspaces.current->name;
 
 	float scale = output->wlr_output->scale;
 	int w = theme->osd_window_switcher_width;
@@ -350,7 +351,7 @@ display_osd(struct output *output, struct wl_array *views)
 	if (output->osd_buffer) {
 		wlr_buffer_drop(&output->osd_buffer->base);
 	}
-	output->osd_buffer = buffer_create_cairo(w, h, scale, true);
+	output->osd_buffer = buffer_create_cairo(w, h, scale);
 	if (!output->osd_buffer) {
 		wlr_log(WLR_ERROR, "Failed to allocate cairo buffer for the window switcher");
 		return;
