@@ -7,10 +7,10 @@
 #include <wlr/backend/headless.h>
 #include <wlr/backend/multi.h>
 #include <wlr/types/wlr_data_control_v1.h>
+#include <wlr/types/wlr_drm.h>
 #include <wlr/types/wlr_export_dmabuf_v1.h>
 #include <wlr/types/wlr_fractional_scale_v1.h>
 #include <wlr/types/wlr_gamma_control_v1.h>
-#include <wlr/types/wlr_input_inhibitor.h>
 #include <wlr/types/wlr_presentation_time.h>
 #include <wlr/types/wlr_primary_selection_v1.h>
 #include <wlr/types/wlr_screencopy_v1.h>
@@ -26,10 +26,15 @@
 #include "config/rcxml.h"
 #include "config/session.h"
 #include "decorations.h"
+#if HAVE_LIBSFDO
+#include "icon-loader.h"
+#endif
 #include "idle.h"
 #include "labwc.h"
 #include "layers.h"
+#include "magnifier.h"
 #include "menu/menu.h"
+#include "output-state.h"
 #include "output-virtual.h"
 #include "regions.h"
 #include "resize-indicator.h"
@@ -40,6 +45,7 @@
 
 #define LAB_WLR_COMPOSITOR_VERSION 5
 #define LAB_WLR_FRACTIONAL_SCALE_V1_VERSION 1
+#define LAB_WLR_LINUX_DMABUF_VERSION 4
 
 static struct wlr_compositor *compositor;
 static struct wl_event_source *sighup_source;
@@ -55,6 +61,11 @@ reload_config_and_theme(struct server *server)
 	theme_finish(server->theme);
 	theme_init(server->theme, server, rc.theme_name);
 
+#if HAVE_LIBSFDO
+	icon_loader_finish(server);
+	icon_loader_init(server);
+#endif
+
 	struct view *view;
 	wl_list_for_each(view, &server->views, link) {
 		view_reload_ssd(view);
@@ -150,61 +161,6 @@ handle_sigchld(int signal, void *data)
 	return 0;
 }
 
-static void
-seat_inhibit_input(struct seat *seat,  struct wl_client *active_client)
-{
-	seat->active_client_while_inhibited = active_client;
-
-	if (seat->focused_layer && active_client !=
-			wl_resource_get_client(seat->focused_layer->resource)) {
-		seat_set_focus_layer(seat, NULL);
-	}
-	struct wlr_surface *previous_kb_surface =
-		seat->seat->keyboard_state.focused_surface;
-	if (previous_kb_surface && active_client !=
-			wl_resource_get_client(previous_kb_surface->resource)) {
-		seat_focus_surface(seat, NULL);	  /* keyboard focus */
-	}
-
-	struct wlr_seat_client *previous_ptr_client =
-		seat->seat->pointer_state.focused_client;
-	if (previous_ptr_client && previous_ptr_client->client != active_client) {
-		wlr_seat_pointer_clear_focus(seat->seat);
-	}
-}
-
-static void
-seat_disinhibit_input(struct seat *seat)
-{
-	seat->active_client_while_inhibited = NULL;
-
-	/*
-	 * Triggers a refocus of the topmost surface layer if necessary
-	 * TODO: Make layer surface focus per-output based on cursor position
-	 */
-	output_update_all_usable_areas(seat->server, /*layout_changed*/ false);
-}
-
-static void
-handle_input_inhibit(struct wl_listener *listener, void *data)
-{
-	wlr_log(WLR_INFO, "activate input inhibit");
-
-	struct server *server =
-		wl_container_of(listener, server, input_inhibit_activate);
-	seat_inhibit_input(&server->seat, server->input_inhibit->active_client);
-}
-
-static void
-handle_input_disinhibit(struct wl_listener *listener, void *data)
-{
-	wlr_log(WLR_INFO, "deactivate input inhibit");
-
-	struct server *server =
-		wl_container_of(listener, server, input_inhibit_deactivate);
-	seat_disinhibit_input(&server->seat);
-}
-
 static void
 handle_drm_lease_request(struct wl_listener *listener, void *data)
 {
@@ -304,6 +260,51 @@ get_headless_backend(struct wlr_backend *backend, void *data)
 	}
 }
 
+static void
+handle_renderer_lost(struct wl_listener *listener, void *data)
+{
+	struct server *server = wl_container_of(listener, server, renderer_lost);
+
+	wlr_log(WLR_INFO, "Re-creating renderer after GPU reset");
+
+	struct wlr_renderer *renderer = wlr_renderer_autocreate(server->backend);
+	if (!renderer) {
+		wlr_log(WLR_ERROR, "Unable to create renderer");
+		return;
+	}
+
+	struct wlr_allocator *allocator =
+		wlr_allocator_autocreate(server->backend, renderer);
+	if (!allocator) {
+		wlr_log(WLR_ERROR, "Unable to create allocator");
+		wlr_renderer_destroy(renderer);
+		return;
+	}
+
+	struct wlr_renderer *old_renderer = server->renderer;
+	struct wlr_allocator *old_allocator = server->allocator;
+	server->renderer = renderer;
+	server->allocator = allocator;
+
+	wl_list_remove(&server->renderer_lost.link);
+	wl_signal_add(&server->renderer->events.lost, &server->renderer_lost);
+
+	wlr_compositor_set_renderer(compositor, renderer);
+
+	struct output *output;
+	wl_list_for_each(output, &server->outputs, link) {
+		wlr_output_init_render(output->wlr_output,
+			server->allocator, server->renderer);
+	}
+
+	reload_config_and_theme(server);
+
+	magnify_reset();
+
+	wlr_allocator_destroy(old_allocator);
+	wlr_renderer_destroy(old_renderer);
+}
+
 void
 server_init(struct server *server)
 {
@@ -346,7 +347,7 @@ server_init(struct server *server)
 	 * window if an x11 server is running.
 	 */
 	server->backend = wlr_backend_autocreate(
-		server->wl_display, &server->session);
+		server->wl_event_loop, &server->session);
 	if (!server->backend) {
 		wlr_log(WLR_ERROR, "unable to create backend");
 		fprintf(stderr, helpful_seat_error_message);
@@ -359,7 +360,8 @@ server_init(struct server *server)
 
 	if (!server->headless.backend) {
 		wlr_log(WLR_DEBUG, "manually creating headless backend");
-		server->headless.backend = wlr_headless_backend_create(server->wl_display);
+		server->headless.backend = wlr_headless_backend_create(
+			server->wl_event_loop);
 	} else {
 		wlr_log(WLR_DEBUG, "headless backend already exists");
 	}
@@ -390,7 +392,26 @@ server_init(struct server *server)
 		exit(EXIT_FAILURE);
 	}
 
-	wlr_renderer_init_wl_display(server->renderer, server->wl_display);
+	server->renderer_lost.notify = handle_renderer_lost;
+	wl_signal_add(&server->renderer->events.lost, &server->renderer_lost);
+
+	if (!wlr_renderer_init_wl_shm(server->renderer, server->wl_display)) {
+		wlr_log(WLR_ERROR, "Failed to initialize shared memory pool");
+		exit(EXIT_FAILURE);
+	}
+
+	if (wlr_renderer_get_texture_formats(
+			server->renderer, WLR_BUFFER_CAP_DMABUF)) {
+		if (wlr_renderer_get_drm_fd(server->renderer) >= 0) {
+			wlr_drm_create(server->wl_display, server->renderer);
+		}
+		server->linux_dmabuf = wlr_linux_dmabuf_v1_create_with_renderer(
+			server->wl_display,
+			LAB_WLR_LINUX_DMABUF_VERSION,
+			server->renderer);
+	} else {
+		wlr_log(WLR_DEBUG, "unable to initialize dmabuf");
+	}
 
 	/*
 	 * Autocreates an allocator for us. The allocator is the bridge between
@@ -499,7 +520,9 @@ server_init(struct server *server)
 		wlr_log(WLR_ERROR, "unable to create presentation interface");
 		exit(EXIT_FAILURE);
 	}
-	wlr_scene_set_presentation(server->scene, presentation);
+	if (server->linux_dmabuf) {
+		wlr_scene_set_linux_dmabuf_v1(server->scene, server->linux_dmabuf);
+	}
 
 	wlr_export_dmabuf_manager_v1_create(server->wl_display);
 	wlr_screencopy_manager_v1_create(server->wl_display);
@@ -522,21 +545,6 @@ server_init(struct server *server)
 	wl_signal_add(&server->constraints->events.new_constraint,
 		&server->new_constraint);
 
-	server->input_inhibit =
-		wlr_input_inhibit_manager_create(server->wl_display);
-	if (!server->input_inhibit) {
-		wlr_log(WLR_ERROR, "unable to create input inhibit manager");
-		exit(EXIT_FAILURE);
-	}
-
-	wl_signal_add(&server->input_inhibit->events.activate,
-		&server->input_inhibit_activate);
-	server->input_inhibit_activate.notify = handle_input_inhibit;
-
-	wl_signal_add(&server->input_inhibit->events.deactivate,
-		&server->input_inhibit_deactivate);
-	server->input_inhibit_deactivate.notify = handle_input_disinhibit;
-
 	server->foreign_toplevel_manager =
 		wlr_foreign_toplevel_manager_v1_create(server->wl_display);
 
@@ -568,6 +576,10 @@ server_init(struct server *server)
 
 	layers_init(server);
 
+#if HAVE_LIBSFDO
+	icon_loader_init(server);
+#endif
+
 #if HAVE_XWAYLAND
 	xwayland_server_init(server, compositor);
 #endif
@@ -614,10 +626,12 @@ server_finish(struct server *server)
 	wl_display_destroy_clients(server->wl_display);
 
 	seat_finish(server);
-	wlr_output_layout_destroy(server->output_layout);
-
 	wl_display_destroy(server->wl_display);
 
 	/* TODO: clean up various scene_tree nodes */
 	workspaces_destroy(server);
+
+#if HAVE_LIBSFDO
+	icon_loader_finish(server);
+#endif
 }
