@@ -22,6 +22,7 @@ ssd_border_create(struct ssd *ssd)
 	int width = view->current.width;
 	int height = view_effective_height(view, /* use_pending */ false);
 	int full_width = width + 2 * theme->border_width;
+	int corner_width = ssd_get_corner_width();
 
 	float *color;
 	struct wlr_scene_tree *parent;
@@ -48,8 +49,8 @@ ssd_border_create(struct ssd *ssd)
 		add_scene_rect(&subtree->parts, LAB_SSD_PART_BOTTOM, parent,
 			full_width, theme->border_width, 0, height, color);
 		add_scene_rect(&subtree->parts, LAB_SSD_PART_TOP, parent,
-			width - 2 * SSD_BUTTON_WIDTH, theme->border_width,
-			theme->border_width + SSD_BUTTON_WIDTH,
+			width - 2 * corner_width, theme->border_width,
+			theme->border_width + corner_width,
 			-(ssd->titlebar.height + theme->border_width), color);
 	} FOR_EACH_END
 
@@ -93,6 +94,7 @@ ssd_border_update(struct ssd *ssd)
 	int width = view->current.width;
 	int height = view_effective_height(view, /* use_pending */ false);
 	int full_width = width + 2 * theme->border_width;
+	int corner_width = ssd_get_corner_width();
 
 	/*
 	 * From here on we have to cover the following border scenarios:
@@ -113,18 +115,18 @@ ssd_border_update(struct ssd *ssd)
 	 *  |_______________|
 	 */
 
-	int side_height = ssd->state.was_tiled_not_maximized
+	int side_height = ssd->state.was_squared
 		? height + ssd->titlebar.height
 		: height;
-	int side_y = ssd->state.was_tiled_not_maximized
+	int side_y = ssd->state.was_squared
 		? -ssd->titlebar.height
 		: 0;
-	int top_width = ssd->titlebar.height <= 0 || ssd->state.was_tiled_not_maximized
+	int top_width = ssd->titlebar.height <= 0 || ssd->state.was_squared
 		? full_width
-		: width - 2 * SSD_BUTTON_WIDTH;
-	int top_x = ssd->titlebar.height <= 0 || ssd->state.was_tiled_not_maximized
+		: width - 2 * corner_width;
+	int top_x = ssd->titlebar.height <= 0 || ssd->state.was_squared
 		? 0
-		: theme->border_width + SSD_BUTTON_WIDTH;
+		: theme->border_width + corner_width;
 
 	struct ssd_part *part;
 	struct wlr_scene_rect *rect;
