@@ -15,10 +15,13 @@
 #include "common/mem.h"
 #include "input/keyboard.h"
 #include "labwc.h"
+#include "protocols/cosmic-workspaces.h"
 #include "view.h"
 #include "workspaces.h"
 #include "xwayland.h"
 
+#define COSMIC_WORKSPACES_VERSION 1
+
 /* Internal helpers */
 static size_t
 parse_workspace_index(const char *name)
@@ -66,7 +69,7 @@ _osd_update(struct server *server)
 		theme->osd_workspace_switcher_boxes_height == 0;
 
 	/* Dimensions */
-	size_t workspace_count = wl_list_length(&server->workspaces);
+	size_t workspace_count = wl_list_length(&server->workspaces.all);
 	uint16_t marker_width = workspace_count * (rect_width + padding) - padding;
 	uint16_t width = margin * 2 + (marker_width < 200 ? 200 : marker_width);
 	uint16_t height = margin * (hide_boxes ? 2 : 3) + rect_height + font_height(&rc.font_osd);
@@ -81,7 +84,7 @@ _osd_update(struct server *server)
 			continue;
 		}
 		struct lab_data_buffer *buffer = buffer_create_cairo(width, height,
-			output->wlr_output->scale, true);
+			output->wlr_output->scale);
 		if (!buffer) {
 			wlr_log(WLR_ERROR, "Failed to allocate buffer for workspace OSD");
 			continue;
@@ -106,8 +109,8 @@ _osd_update(struct server *server)
 		uint16_t x;
 		if (!hide_boxes) {
 			x = (width - marker_width) / 2;
-			wl_list_for_each(workspace, &server->workspaces, link) {
-				bool active =  workspace == server->workspace_current;
+			wl_list_for_each(workspace, &server->workspaces.all, link) {
+				bool active =  workspace == server->workspaces.current;
 				set_cairo_color(cairo, server->theme->osd_label_text_color);
 				cairo_rectangle(cairo, x, margin,
 					rect_width - padding, rect_height);
@@ -124,10 +127,11 @@ _osd_update(struct server *server)
 		/* Text */
 		set_cairo_color(cairo, server->theme->osd_label_text_color);
 		PangoLayout *layout = pango_cairo_create_layout(cairo);
+		pango_context_set_round_glyph_positions(pango_layout_get_context(layout), false);
 		pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_END);
 
 		/* Center workspace indicator on the x axis */
-		int req_width = font_width(&rc.font_osd, server->workspace_current->name);
+		int req_width = font_width(&rc.font_osd, server->workspaces.current->name);
 		req_width = MIN(req_width, width - 2 * margin);
 		x = (width - req_width) / 2;
 		if (!hide_boxes) {
@@ -140,7 +144,7 @@ _osd_update(struct server *server)
 		pango_layout_set_font_description(layout, desc);
 		pango_layout_set_width(layout, req_width * PANGO_SCALE);
 		pango_font_description_free(desc);
-		pango_layout_set_text(layout, server->workspace_current->name, -1);
+		pango_layout_set_text(layout, server->workspaces.current->name, -1);
 		pango_cairo_show_layout(cairo, layout);
 
 		g_object_unref(layout);
@@ -164,13 +168,22 @@ _osd_update(struct server *server)
 		wlr_scene_node_set_position(&output->workspace_osd->node, lx, ly);
 		wlr_scene_buffer_set_buffer(output->workspace_osd, &buffer->base);
 		wlr_scene_buffer_set_dest_size(output->workspace_osd,
-			buffer->unscaled_width, buffer->unscaled_height);
+			buffer->logical_width, buffer->logical_height);
 
 		/* And finally drop the buffer so it will get destroyed on OSD hide */
 		wlr_buffer_drop(&buffer->base);
 	}
 }
 
+/* cosmic workspace handlers */
+static void
+handle_workspace_activate(struct wl_listener *listener, void *data)
+{
+	struct workspace *workspace = wl_container_of(listener, workspace, on.activate);
+	workspaces_switch_to(workspace, /* update_focus */ true);
+	wlr_log(WLR_INFO, "activating workspace %s", workspace->name);
+}
+
 /* Internal API */
 static void
 add_workspace(struct server *server, const char *name)
@@ -179,12 +192,21 @@ add_workspace(struct server *server, const char *name)
 	workspace->server = server;
 	workspace->name = xstrdup(name);
 	workspace->tree = wlr_scene_tree_create(server->view_tree);
-	wl_list_append(&server->workspaces, &workspace->link);
-	if (!server->workspace_current) {
-		server->workspace_current = workspace;
+	wl_list_append(&server->workspaces.all, &workspace->link);
+	if (!server->workspaces.current) {
+		server->workspaces.current = workspace;
 	} else {
 		wlr_scene_node_set_enabled(&workspace->tree->node, false);
 	}
+
+	bool active = server->workspaces.current == workspace;
+	workspace->cosmic_workspace = lab_cosmic_workspace_create(server->workspaces.cosmic_group);
+	lab_cosmic_workspace_set_name(workspace->cosmic_workspace, name);
+	lab_cosmic_workspace_set_active(workspace->cosmic_workspace, active);
+	lab_cosmic_workspace_set_hidden(workspace->cosmic_workspace, !active);
+
+	workspace->on.activate.notify = handle_workspace_activate;
+	wl_signal_add(&workspace->cosmic_workspace->events.activate, &workspace->on.activate);
 }
 
 static struct workspace *
@@ -259,7 +281,14 @@ _osd_show(struct server *server)
 void
 workspaces_init(struct server *server)
 {
-	wl_list_init(&server->workspaces);
+	server->workspaces.cosmic_manager = lab_cosmic_workspace_manager_create(
+		server->wl_display, /* capabilities */ CW_CAP_WS_ACTIVATE,
+		COSMIC_WORKSPACES_VERSION);
+
+	server->workspaces.cosmic_group = lab_cosmic_workspace_group_create(
+		server->workspaces.cosmic_manager);
+
+	wl_list_init(&server->workspaces.all);
 
 	struct workspace *conf;
 	wl_list_for_each(conf, &rc.workspace_config.workspaces, link) {
@@ -277,13 +306,18 @@ workspaces_switch_to(struct workspace *target, bool update_focus)
 {
 	assert(target);
 	struct server *server = target->server;
-	if (target == server->workspace_current) {
+	if (target == server->workspaces.current) {
 		return;
 	}
 
 	/* Disable the old workspace */
 	wlr_scene_node_set_enabled(
-		&server->workspace_current->tree->node, false);
+		&server->workspaces.current->tree->node, false);
+
+	lab_cosmic_workspace_set_active(
+		server->workspaces.current->cosmic_workspace, false);
+	lab_cosmic_workspace_set_hidden(
+		server->workspaces.current->cosmic_workspace, true);
 
 	/* Move Omnipresent views to new workspace */
 	struct view *view;
@@ -299,10 +333,10 @@ workspaces_switch_to(struct workspace *target, bool update_focus)
 	wlr_scene_node_set_enabled(&target->tree->node, true);
 
 	/* Save the last visited workspace */
-	server->workspace_last = server->workspace_current;
+	server->workspaces.last = server->workspaces.current;
 
 	/* Make sure new views will spawn on the new workspace */
-	server->workspace_current = target;
+	server->workspaces.current = target;
 
 	/*
 	 * Make sure we are focusing what the user sees.
@@ -330,6 +364,9 @@ workspaces_switch_to(struct workspace *target, bool update_focus)
 
 	/* Ensure that only currently visible fullscreen windows hide the top layer */
 	desktop_update_top_layer_visiblity(server);
+
+	lab_cosmic_workspace_set_active(target->cosmic_workspace, true);
+	lab_cosmic_workspace_set_hidden(target->cosmic_workspace, false);
 }
 
 void
@@ -361,7 +398,7 @@ workspaces_find(struct workspace *anchor, const char *name, bool wrap)
 	size_t index = 0;
 	struct workspace *target;
 	size_t wants_index = parse_workspace_index(name);
-	struct wl_list *workspaces = &anchor->server->workspaces;
+	struct wl_list *workspaces = &anchor->server->workspaces.all;
 
 	if (wants_index) {
 		wl_list_for_each(target, workspaces, link) {
@@ -372,7 +409,7 @@ workspaces_find(struct workspace *anchor, const char *name, bool wrap)
 	} else if (!strcasecmp(name, "current")) {
 		return anchor;
 	} else if (!strcasecmp(name, "last")) {
-		return anchor->server->workspace_last;
+		return anchor->server->workspaces.last;
 	} else if (!strcasecmp(name, "left")) {
 		return get_prev(anchor, workspaces, wrap);
 	} else if (!strcasecmp(name, "right")) {
@@ -394,6 +431,8 @@ destroy_workspace(struct workspace *workspace)
 	wlr_scene_node_destroy(&workspace->tree->node);
 	zfree(workspace->name);
 	wl_list_remove(&workspace->link);
+
+	lab_cosmic_workspace_destroy(workspace->cosmic_workspace);
 	free(workspace);
 }
 
@@ -407,7 +446,7 @@ workspaces_reconfigure(struct server *server)
 	 *   - Destroy workspaces if fewer workspace are desired
 	 */
 
-	struct wl_list *actual_workspace_link = server->workspaces.next;
+	struct wl_list *actual_workspace_link = server->workspaces.all.next;
 
 	struct workspace *configured_workspace;
 	wl_list_for_each(configured_workspace,
@@ -415,7 +454,7 @@ workspaces_reconfigure(struct server *server)
 		struct workspace *actual_workspace = wl_container_of(
 			actual_workspace_link, actual_workspace, link);
 
-		if (actual_workspace_link == &server->workspaces) {
+		if (actual_workspace_link == &server->workspaces.all) {
 			/* # of configured workspaces increased */
 			wlr_log(WLR_DEBUG, "Adding workspace \"%s\"",
 				configured_workspace->name);
@@ -428,20 +467,22 @@ workspaces_reconfigure(struct server *server)
 				actual_workspace->name, configured_workspace->name);
 			free(actual_workspace->name);
 			actual_workspace->name = xstrdup(configured_workspace->name);
+			lab_cosmic_workspace_set_name(
+				actual_workspace->cosmic_workspace, actual_workspace->name);
 		}
 		actual_workspace_link = actual_workspace_link->next;
 	}
 
-	if (actual_workspace_link == &server->workspaces) {
+	if (actual_workspace_link == &server->workspaces.all) {
 		return;
 	}
 
 	/* # of configured workspaces decreased */
 	overlay_hide(&server->seat);
 	struct workspace *first_workspace =
-		wl_container_of(server->workspaces.next, first_workspace, link);
+		wl_container_of(server->workspaces.all.next, first_workspace, link);
 
-	while (actual_workspace_link != &server->workspaces) {
+	while (actual_workspace_link != &server->workspaces.all) {
 		struct workspace *actual_workspace = wl_container_of(
 			actual_workspace_link, actual_workspace, link);
 
@@ -455,12 +496,12 @@ workspaces_reconfigure(struct server *server)
 			}
 		}
 
-		if (server->workspace_current == actual_workspace) {
+		if (server->workspaces.current == actual_workspace) {
 			workspaces_switch_to(first_workspace,
 				/* update_focus */ true);
 		}
-		if (server->workspace_last == actual_workspace) {
-			server->workspace_last = first_workspace;
+		if (server->workspaces.last == actual_workspace) {
+			server->workspaces.last = first_workspace;
 		}
 
 		actual_workspace_link = actual_workspace_link->next;
@@ -472,8 +513,8 @@ void
 workspaces_destroy(struct server *server)
 {
 	struct workspace *workspace, *tmp;
-	wl_list_for_each_safe(workspace, tmp, &server->workspaces, link) {
+	wl_list_for_each_safe(workspace, tmp, &server->workspaces.all, link) {
 		destroy_workspace(workspace);
 	}
-	assert(wl_list_empty(&server->workspaces));
+	assert(wl_list_empty(&server->workspaces.all));
 }
