@@ -7,6 +7,7 @@
  *	- keeping non-layer-shell xdg-popups outside the layers.c code
  */
 
+#include "common/macros.h"
 #include "common/mem.h"
 #include "labwc.h"
 #include "node.h"
@@ -27,15 +28,41 @@ popup_unconstrain(struct xdg_popup *popup)
 {
 	struct view *view = popup->parent_view;
 	struct server *server = view->server;
-	struct wlr_box *popup_box = &popup->wlr_popup->current.geometry;
+
+	/* Get position of parent toplevel/popup */
+	int parent_lx, parent_ly;
+	struct wlr_scene_tree *parent_tree = popup->wlr_popup->parent->data;
+	wlr_scene_node_coords(&parent_tree->node, &parent_lx, &parent_ly);
+
+	/*
+	 * Get usable area to constrain by
+	 *
+	 * The scheduled top-left corner (x, y) of the popup is sometimes less
+	 * than zero, typically with Qt apps. We therefore clamp it to avoid for
+	 * example the 'File' menu of a maximized window to end up on an another
+	 * output.
+	 */
+	struct wlr_box *popup_box = &popup->wlr_popup->scheduled.geometry;
 	struct output *output = output_nearest_to(server,
-		view->current.x + popup_box->x,
-		view->current.y + popup_box->y);
+		parent_lx + MAX(popup_box->x, 0),
+		parent_ly + MAX(popup_box->y, 0));
 	struct wlr_box usable = output_usable_area_in_layout_coords(output);
 
+	/* Get offset of toplevel window from its surface */
+	int toplevel_dx = 0;
+	int toplevel_dy = 0;
+	struct wlr_xdg_surface *toplevel_surface = xdg_surface_from_view(view);
+	if (toplevel_surface) {
+		toplevel_dx = toplevel_surface->current.geometry.x;
+		toplevel_dy = toplevel_surface->current.geometry.y;
+	} else {
+		wlr_log(WLR_ERROR, "toplevel is not valid XDG surface");
+	}
+
+	/* Geometry of usable area relative to toplevel surface */
 	struct wlr_box output_toplevel_box = {
-		.x = usable.x - view->current.x,
-		.y = usable.y - view->current.y,
+		.x = usable.x - (view->current.x - toplevel_dx),
+		.y = usable.y - (view->current.y - toplevel_dy),
 		.width = usable.width,
 		.height = usable.height,
 	};
@@ -101,7 +128,7 @@ xdg_popup_create(struct view *view, struct wlr_xdg_popup *wlr_popup)
 	popup->wlr_popup = wlr_popup;
 
 	popup->destroy.notify = handle_xdg_popup_destroy;
-	wl_signal_add(&wlr_popup->base->events.destroy, &popup->destroy);
+	wl_signal_add(&wlr_popup->events.destroy, &popup->destroy);
 
 	popup->new_popup.notify = popup_handle_new_xdg_popup;
 	wl_signal_add(&wlr_popup->base->events.new_popup, &popup->new_popup);
@@ -116,7 +143,7 @@ xdg_popup_create(struct view *view, struct wlr_xdg_popup *wlr_popup)
 	 * We must add xdg popups to the scene graph so they get rendered. The
 	 * wlroots scene graph provides a helper for this, but to use it we must
 	 * provide the proper parent scene node of the xdg popup. To enable
-	 * this, we always set the user data field of xdg_surfaces to the
+	 * this, we always set the user data field of wlr_surfaces to the
 	 * corresponding scene node.
 	 *
 	 * xdg-popups live in server->xdg_popup_tree so that they can be
